fl

{
Something like Gimbal by using Servo SG90 and L3GD20
PropForth5.5


2015/03/21 11:18:21


Servo curcuit     

                 3.3V       5V                5V
   MOSFET(2N7000) |         | RED              |
           -----------    10kohm               |RED
          |       G   |     |    orange   -----------
  P0  ----|S 2N7000  D|------------------|PWM   SG90 | Rotate
           -----------                    -----------
                                               |BROWN
                                              GND

                 3.3V       5V                 5V
   MOSFET(2N7000) |         |                  |
           -----------    10kohm               |RED
          |       G   |     |    orange   -----------
  P1  ----|S 2N7000  D|------------------|PWM   SG90 | Up/Down
           -----------                    -----------
                                               |Brown
                                              GND 
                                              
                                              
L3GD20 module      Propeller
       1 VDD    ----  3.3V
       2 SCL    ----  P28   
       3 SDA    ----  P29 
       4 SA0    ----  3.3V
       5 CS     ----  3.3V
       8 GND    ----  GND
                                              
}


\ =========================================================================== 
\ Constants 
\ =========================================================================== 
\ --- Jiro sensor ---
\ Slave addres h6B for L3GD20 SA0=1
hD6 wconstant L3GD20

\ register name
h20 wconstant ctrl1
h23 wconstant ctrl4
h27 wconstant status
h28 wconstant out_x
h2A wconstant out_y

\ --- Servo --- 
\ Special register
h1F8	wconstant ctra
h1FA	wconstant frqa 
h1FC	wconstant phsa 
\ Time for servo(SG90)
d80 wconstant 1usec
d1000 wconstant SV_MIN
d2000 wconstant SV_MAX
\ channel[max:8]
2 wconstant ch
\ top pin number of contiguous servo group
0 wconstant pin

\ =========================================================================== 
\ Variables 
\ =========================================================================== 
\ --- Jyro sensor ---
variable cx
variable cy
variable x_vel
variable y_vel
variable x_angle
variable y_angle
variable x_prev
variable y_prev

\ --- Servo --- 
\ target[ch],delta[ch],usec[ch]: each channel has 2byte
variable target 
variable delta 
variable usec 

\ =========================================================================== 
\ Main
\ =========================================================================== 
\ --- Jyro sensor -----------------------------------------------------------------------
\ Convert 2's complement form(16bit) to 32bit
\ None if positive
\ ( n1 -- n2 ) n1:2's complement form(16bit)  n2:32bit
: conv_32 dup h8000 and if hFFFF0000 or then ;

\ Wait until data is ready
\ ( -- )          
: dataReady?
begin
     1 status L3GD20 i2c_rd_multi err?  \ Read status register 
     8 and 0=                           \ Check bit3
     if 0 else 1 then
until                                   
;

\ Get x,y
\ ( -- x y )
: get_xy
\ Get 4byte  
4 out_x h80 or                               \ In case of multi_read, register set bit7 to 1
L3GD20 i2c_rd_multi err?                     \ Get x,y    
8 lshift or conv_32 >r                       \ y
8 lshift or conv_32                          \ x
r> 
;

\ Add n1 to the long contents of address
\ ( n1 addr -- )
: L+! dup L@ rot + swap L! ;

\ Get offset for x,y
\ ( -- )
: offset
0 cx L! 0 cy L!  
d100 0 do
     dataReady?
     get_xy                    
     cy L+! cx L+!
loop
cx L@ d100 / cx L!
cy L@ d100 / cy L!
;

\ --- Servo ------------------------------------------------------------------------------
\ target position for each ch
\ ( n1 -- n2 )  n1:ch n2:target_position's varriable address for each ch
: tgt_pos 2 u* target + ;

\ delta position for each ch
\ ( n1 -- n2 )  n1:ch n2:delta's varriable address for each ch
: delta_val 2 u* delta + ;

\ current position for each ch
\ ( n1 -- n2 )  n1:ch n2:current_position's varriable address for each ch
: cur_pos 2 u* usec + ;

\ Set servo-motor's pin to output
\ ( n1 -- )   n1:top pin number of contiguous servo group
: setup ch 0 do dup pinout 1+ loop drop ;

\ Check if position is inside region
\ ( n1 -- n2 )  
\ n1:servo position(microsecond)  
\ n2:SV_MIN <= n2 <= SV_MAX
: pos_check SV_MIN W@ max SV_MAX W@ min ;

\ Set selected servo(ch) to new position 
\ ( n1 n2 -- )  
\ n1:ch [0,1,2,3,4,5,6,7]
\ n2:newpos specified in microseconds
: setpos
pos_check                       
swap 2dup                                    \ ( n2 n1 n2 n1 )
tgt_pos W!                                   \ ( n2 n1 )   Save target_position to sprcified ch
dup                                          \ ( n2 n1 n1 )
\ delta[ch] = 0
delta_val 0 swap W!                          \ ( n2 n1 )  Save delta=0 to specified ch
\ usec[ch] = newpos
cur_pos W!                                   \ ( -- )  Save new_position to specified ch
;

\ Move selected servo from current position to newpos in n1 frames
\ ( n1 n2 n3 -- )  
\ n1:frame [1frame is 20msec(servo frame timing)]
\ n2:ch
\ n3:newpos
: movpos
pos_check
2dup                               \ ( n1 n2 n3 n2 n3 )
\ save newpos to target[ch]
swap tgt_pos W!                    \ ( n1 n2 n3 )  Save target_position to sprcified ch
rot                                \ ( n2 n3 n1 )
dup 1 =
if
     drop                          \ ( n2 n3 )
     setpos
else                                   
     >r                            \ ( n2 n3 )   Push frame
     over                          \ ( n2 n3 n2 )
     cur_pos W@ -                  \ Calculate (difference from newpos and usec)
     abs d10 u*                    \ ( n2 n3 calc_value )
     r>                            \ Pop frame
     u/                            \ ( n2 n3 calc_value )  Calculate 0.1usec/frame
     5 + d10 u/ 1 max              \ ( n2 compared_result )  round up to 1usec/frame
     swap                          \ ( compared_resulte n2 )
     delta_val W!                  \ ( -- )  Update delta to specified ch
then
;

\ Returns position (current output command to servo) of channel
\ ( n1 -- n2 )  
\ n1:ch  n2:ch's position(microseconds)
: position cur_pos W@ ;

\ Returns True when channel is at desired target
\ ( n1 -- n2 )  
\ n1:ch  n2:t/f
: at_target dup cur_pos W@ swap tgt_pos W@ = ;

\ Wait until servo channel reaches at target position
\ ( n1 -- )  n1:ch
: wait_tgt begin dup at_target until drop ;

\ Runs 2 servos
\ ( n1 -- )  n1:top pin number of contiguous servo group
: delta_servo
dup 
setup
1 frqa COG!
0 phsa COG!
cnt COG@ d200000 +            \ cnt + 2.5msec  (2.5msec x 8 = 20msec)
swap
begin
     \ run 8 slot(20msec)
     8 0 do   
          \ active channel?            
          i ch <                           
          if                                
               \ Set PWM/NCO mode on servo pin       
               dup i + h10000000 or ctra COG!
               \ Set negative value to phsa      
               i cur_pos W@ 1usec u* negate phsa COG!
               
               i cur_pos W@ i tgt_pos W@ <
               if
                    i cur_pos W@ i delta_val W@ +
                    i tgt_pos W@ min
                    i cur_pos W!
               else
                    i cur_pos W@ i tgt_pos W@ >
                    if
                         i cur_pos W@ i delta_val W@ -
                         i tgt_pos W@ max
                         i cur_pos W!
                    then
               then
          then 
          swap                                                    
          d200000 waitcnt                         \ cnt + 2.5msec
          swap
          0 ctra COG! 
     loop
0 until
\ fkey? swap drop until drop
;

\ Start servo cog
\ ( -- )   
: servo_init
ch 0 do   
     i d1400 setpos                \ servo1 initial position[0degree]      
     i d1400 setpos                \ servo2 initial position[0degree]
loop
 c" pin delta_servo" 0 cogx
;

: disp_pos
ch 0 do 
     ." ch:" i .
     ." target:" i tgt_pos W@ .
     ." delta:" i delta_val W@ .
     ." usec:" i cur_pos W@ .
     cr
loop
cr
;

\ ----------------------------------------------------------------------------------------
\  Main
\ ----------------------------------------------------------------------------------------

\ Get angle-velocity[x,y] during 10ms
\ ( -- )
: angle_vel
\ init L3GD20
hCF 1 ctrl1 L3GD20 i2c_wr_multi err?            
h00 1 ctrl4 L3GD20 i2c_wr_multi err?
d1000 delms
offset

d800000 cnt COG@ +    \ 10msec        
begin                    
     0 0                                \ initial x and y  ( nx ny )
     \ Get x and y 10times
     d10 0 do
          get_xy                        \ ( nx ny x y )
          rot + swap rot + swap         \ ( nx+x ny+y )
     loop
     \ Subtract offset value
     cy L@ d10 * - d10 / y_vel L!
     cx L@ d10 * - d10 / x_vel L!                 
                                      
     d800000 waitcnt              
\     fkey? swap drop
0 until 
\ drop                  
;



: gimbal
c" angle_vel" 1 cogx
d2000 delms

\ Init servo
servo_init
d1000 delms

0 x_angle L!
0 x_prev L!                                        
0 y_angle L!
0 y_prev L!                                         
  
d800000 cnt COG@ +    \ 10msec 
begin                    
     d10 0 do
          x_prev L@ x_vel L@ dup x_prev L!
          + d114 / 2 /                            \ ( cnt [x_prev+x_current]/114/2 )
          x_angle L+!                             \ ( [cnt )
          y_prev L@ y_vel L@ dup y_prev L!
          + d114 / 2 /                            \ ( cnt [y_prev+y_current]/114/2 )
          y_angle L+!                             \ ( [cnt )          
          d800000 waitcnt
     loop
     \ ch0:x-axis       
     0 
     x_angle L@ d100 / d500 * d50 / d1400 + setpos           
     \ ch1:y-axis 
     1
     y_angle L@ d100 / d500 * d48 / d1400 + setpos
     fkey? swap drop
 until 
drop
0 cogreset 
1 cogreset                
;
